/******************************************************************************
  Copyright (c) 1992 Xerox Corporation.  All rights reserved.
  Portions of this code were written by Stephen White, aka ghond.
  Use and copying of this software and preparation of derivative works based
  upon this software are permitted.  Any distribution of this software or
  derivative works must comply with all applicable United States export
  control laws.  This software is made available AS IS, and Xerox Corporation
  makes no warranty about the software, its performance or its conformity to
  any specification.  Any person obtaining a copy of this software is requested
  to send their name and post office or electronic mail address to:
    Pavel Curtis
    Xerox PARC
    3333 Coyote Hill Rd.
    Palo Alto, CA 94304
    Pavel@Xerox.Com
 *****************************************************************************/

#include "ast.h"
#include "bytecodes.h"
#include "config.h"
#include "db.h"
#include "eval_env.h"
#include "eval_vm.h"
#include "execute.h"
#include "list.h"
#include "log.h"
#include "opcode.h"
#include "options.h"
#include "server.h"
#include "storage.h"
#include "streams.h"
#include "structures.h"
#include "tasks.h"
#include "timers.h"
#include "unparse.h"
#include "utils.h"

/* the following globals are the guts of the virtual machine: */
static activation      *activ_stack = 0;
static unsigned		max_stack_size = 0;
static unsigned    	top_activ_stack; /* points to top-of-stack
					    (last-occupied-slot),
					    not next-empty-slot */
static int		root_activ_vector; /* root_activ_vector == MAIN_VECTOR
					      iff root activation is main
					      vector */

/* these globals are not part of the vm because they get re-initialized
   after a suspend */
static int         ticks_remaining;
int		   task_timed_out;
static int	   interpreter_is_running = 0;
static int         aborting_task;
static Timer_ID	   task_alarm_id;
static task_kind   current_task_kind; 

/* macros to ease indexing into activation stack */
#define RUN_ACTIV     activ_stack[top_activ_stack]
#define CALLER_ACTIV  activ_stack[top_activ_stack - 1]

/**** error handling ****/

static void
print_backtrace_line(const char *line, int to_log)
{
    if (to_log)
	logf("%s\n", line);
    else
	notify(RUN_ACTIV.player, line);
}

void
print_error_backtrace(const char *msg, int to_log)
{
    int	        t;
    Stream     *str;

    if (!interpreter_is_running)
	return;
    str = new_stream(100);
    for (t = top_activ_stack; t >= 0; t--) {
	if (t != top_activ_stack)
	    stream_printf(str, "... called from ");
	stream_printf(str, "#%d:%s", activ_stack[t].vloc,
		                     activ_stack[t].verbname);
	if (activ_stack[t].vloc != activ_stack[t].this)
	    stream_printf(str, " (this == #%d)", activ_stack[t].this);

	stream_printf(str, ", line %d",
		      get_lineno(activ_stack[t].prog, activ_stack[t].error_pc,
				 (t == 0) ? root_activ_vector : MAIN_VECTOR));
	if (t == top_activ_stack)
	    stream_printf(str, ":  %s", msg);
	print_backtrace_line(reset_stream(str), to_log);
	if (t > 0 && activ_stack[t].bi_func_pc) {
	    stream_printf(str, "... called from built-in function %s()",
			  name_func_by_num(activ_stack[t].bi_func_id));
	    print_backtrace_line(reset_stream(str), to_log);
	}
    }
    print_backtrace_line("(End of traceback)", to_log);
    free_stream(str);
}

static void
abort_task(const char *msg) 
{
    print_error_backtrace(msg, 0);
    aborting_task = 1;
}

/**** activation manipulation ****/

static int
push_activation(void)
{
    if (top_activ_stack < max_stack_size - 1) {
	top_activ_stack++;
	return 1;
    } else
	return 0;
}    

void 
free_activation(activation a, char data_too)
{
    Var *i;
    
    free_rt_env(a.rt_env, a.prog->num_var_names);
    
    for (i = a.base_rt_stack; i < a.top_rt_stack; i++)
	free_var(*i);
    myfree((void *) a.base_rt_stack, M_RT_STACK);
    free_var(a.temp);
    free_str(a.verb);
    free_str(a.verbname);
    
    free_program(a.prog);
    
    if (data_too && a.bi_func_pc && a.bi_func_data)
	free_data(a.bi_func_data);
    /* else bi_func_state will be later freed by bi_function */
}

/**** run time environment setup ****/

void
verb_call_setup_rt_env(char *verb, Objid what, Var args)
{
    /* consumes verb and args */
    
    Var *new = new_rt_env(RUN_ACTIV.prog->num_var_names);
    Var v;
    
    fill_in_rt_consts(new);

    set_rt_env_obj(new, SLOT_THIS, what);
    set_rt_env_obj(new, SLOT_CALLER, CALLER_ACTIV.this);
    
#define ENV_COPY(slot) \
    (set_rt_env_var(new, slot, var_ref(CALLER_ACTIV.rt_env[slot])))
    
    ENV_COPY(SLOT_ARGSTR);
    ENV_COPY(SLOT_DOBJ);
    ENV_COPY(SLOT_DOBJSTR); 
    ENV_COPY(SLOT_PREPSTR); 
    ENV_COPY(SLOT_IOBJ);
    ENV_COPY(SLOT_IOBJSTR);
    
    if (is_wizard(CALLER_ACTIV.progr) &&
	(CALLER_ACTIV.rt_env[SLOT_PLAYER].type == TYPE_OBJ))
	ENV_COPY(SLOT_PLAYER);
    else
	set_rt_env_obj(new, SLOT_PLAYER, CALLER_ACTIV.player);
    RUN_ACTIV.player = new[SLOT_PLAYER].v.obj;
    
    v.type = TYPE_STR;
    v.v.str = verb;
    set_rt_env_var(new, SLOT_VERB, v); 	  /* no var_dup */    
    set_rt_env_var(new, SLOT_ARGS, args); /* no var_dup */
    
    RUN_ACTIV.rt_env = new;
}


/** Set up another activation for calling a verb
  does not change the vm in case of any error **/

static enum error
full_call_verb(Objid this, char *verb, Var args, Objid vloc,
	       const char *verbname,  Program *prog, Objid progr,
	       int debug)
{
    /* verb is consumed */
    if (!push_activation()) {
	free_str(verb);
	return E_MAXREC;
    }
    
    add_ref_to_program(prog);
    RUN_ACTIV.prog = prog;
    RUN_ACTIV.this = this;
    RUN_ACTIV.progr = progr;
    RUN_ACTIV.vloc = vloc;
    RUN_ACTIV.verb= str_ref(verb);
    RUN_ACTIV.verbname = str_ref(verbname);
    RUN_ACTIV.debug = debug;
    verb_call_setup_rt_env(verb, this, args);
    /* verb, args are consumed in the new rt_env.  they can't be freed. */

    RUN_ACTIV.top_rt_stack = RUN_ACTIV.base_rt_stack
	= mymalloc (RUN_ACTIV.prog->main_vector.max_stack * sizeof(Var),
		    M_RT_STACK);
    RUN_ACTIV.pc = 0;
    RUN_ACTIV.error_pc = 0;
    RUN_ACTIV.bi_func_pc = 0;
    RUN_ACTIV.temp.type = TYPE_ANY;

    return E_NONE;
}

enum error 
call_verb(Objid obj, const char *vname, Var args, int do_pass)
{
    /* if call succeeds, args will be consumed.  If call fails, args
       will NOT be consumed  -- it must therefore be freed by caller */
    /* vname will never be consumed */

    /* will only return E_MAXREC, E_INVIND, E_VERBNF, E_TYPE (for args)
       or E_NONE */
    /* returns an error if there is one, and does not change the vm in that
       case, else sets up the activ_stack for the verb call and then returns
       E_NONE */

    Objid  	where = obj;
    verbdef_info v_i;

    if (do_pass) 
	if (!valid(RUN_ACTIV.vloc))
	    return E_INVIND;
	else
	    where = obj_parent(RUN_ACTIV.vloc);

    v_i = get_verbdef_info(where, vname);
    if (!verb_exists(v_i) || !program_from_info(v_i))
	return E_VERBNF;

    return(full_call_verb(obj, str_dup(vname), args, v_i.vloc,
			  name_from_info(v_i), program_from_info(v_i),
			  owner_from_info(v_i),
			  perms_from_info(v_i) & (1 << VP_DEBUG)));
}

static int
rangeset_check(int end, int from, int to)
{
    if (from > end+1 || to < 0)
	return 1;
    return 0;
}

/** 
  the main interpreter -- run()
  everything is just an entry point to it
**/  
	    
static enum outcome
run(enum error resumption_error, Var *result) /* runs the_vm */
{
    /* If the returned value is OUTCOME_DONE and RESULT is non-NULL, then
     * *RESULT is the value returned by the top frame.
     */

    /* bc, bv, rts are distinguished as the state variables of run()
       their value capture the state of the running between OP_ cases */
    Bytecodes bc;
    register Byte *bv, *error_bv;
    register Var  *rts; /* next empty slot */
    register enum Opcode op;
    package p;
    unsigned func_id = 0;
    Var error_var;
    
    /** a bunch of macros that work *ONLY* inside run() **/

/* helping macros about the runtime_stack. */
#define POP()         (*(--rts))
#define PUSH(v)       (*(rts++) = v)
#define PUSH_DUP(v)   PUSH(var_ref(v))
#define TOP_RT_VALUE           (*(rts - 1))
#define NEXT_TOP_RT_VALUE      (*(rts - 2))

#define READ_BYTES(bv, nb)			\
    ( bv += nb,					\
      (nb == 1				        \
       ? bv[-1]					\
       : (nb == 2				\
	  ? ((unsigned) bv[-2] << 8) + bv[-1]	\
	  : (((unsigned) bv[-4] << 24)		\
	     + ((unsigned) bv[-3] << 16)     	\
	     + ((unsigned) bv[-2] << 8)      	\
	     + bv[-1]))))

#define LOAD_STATE_VARIABLES() 					\
{  								\
    bc = ( (top_activ_stack != 0 || root_activ_vector == MAIN_VECTOR) \
	   ? RUN_ACTIV.prog->main_vector 			\
	   : RUN_ACTIV.prog->fork_vectors[root_activ_vector]); 	\
    bv = bc.vector + RUN_ACTIV.pc;  				\
    error_bv = bc.vector + RUN_ACTIV.error_pc;			\
    rts = RUN_ACTIV.top_rt_stack; /* next empty slot */        	\
}

#define STORE_STATE_VARIABLES()			\
{						\
    RUN_ACTIV.pc = bv - bc.vector;		\
    RUN_ACTIV.error_pc = error_bv - bc.vector;	\
    RUN_ACTIV.top_rt_stack = rts; }
    
#define RAISE_ERROR(the_err) 			\
{						\
    error_var.type = TYPE_ERR; 		       	\
    error_var.v.err = the_err; 		       	\
    if (RUN_ACTIV.debug) { 			\
	STORE_STATE_VARIABLES();		\
        abort_task(unparse_error(the_err)); 	\
    } 						\
}

#define PUSH_ERROR(the_err)                                     \
{    							    	\
    RAISE_ERROR(the_err); /* error_var has been prepared by */  \
    PUSH(error_var);  	  /* RAISE_ERROR(), above */ 		\
}

#define JUMP(label)     (bv = bc.vector + label)

/* end of major run() macros */

  LOAD_STATE_VARIABLES();

  if (resumption_error != E_NONE) {
      error_bv = bv;
      PUSH_ERROR(resumption_error);
  }

  for (;;) { 
    op = *bv;
    error_bv = bv;
    bv++;

    if (COUNT_TICK(op))
      ticks_remaining--;
    
    if (!aborting_task) {
	if (ticks_remaining <= 0) {
	    STORE_STATE_VARIABLES();
	    abort_task("Task ran out of ticks");
	}
	
	if (task_timed_out) {
	    STORE_STATE_VARIABLES();
	    abort_task("Task ran out of seconds");
	}
    }

    if (aborting_task)	/* The elegant way to unwind the stack properly... */
	op = OP_DONE;

    switch (op) {

      case OP_EXTENDED:
	{
	    register enum Extended_Opcode op2 = *bv;
	    bv++;
	    switch (op2) {
	      case OP_RANGESET:
		{
		    Var base, from, to, value;		
		    
		    value = POP();	/* rhs value (list or string) */
		    to = POP();		/* end of range (number) */
		    from = POP();	/* start of range (number) */
		    base = POP();	/* lhs (list or string) */
		    /* base[from..to] = value */
		    if (to.type != TYPE_NUM || from.type != TYPE_NUM
			|| (base.type != TYPE_LIST && base.type != TYPE_STR)
			|| (value.type != TYPE_LIST && value.type != TYPE_STR)
			|| (base.type != value.type)) {
			free_var(base);
			free_var(to);
			free_var(from);
			free_var(value);
			PUSH_ERROR(E_TYPE);
		    } else if (rangeset_check(base.type == TYPE_STR
					      ? strlen(base.v.str)
					      : base.v.list[0].v.num,
					      from.v.num, to.v.num)) {
			free_var(base);
			free_var(to);
			free_var(from);
			free_var(value);
			PUSH_ERROR(E_RANGE);
		    } else if (base.type == TYPE_LIST)
			PUSH(listrangeset(base, from.v.num, to.v.num, value));
		    else /* TYPE_STR */
			PUSH(strrangeset(base, from.v.num, to.v.num, value));
		}
		break;
		
	      default: ;

	    }
	}
	break;
	  
      case OP_IF_QUES:
      case OP_IF:
      case OP_WHILE:
      case OP_EIF:
	{
	    Var cond;
	    
	    cond = POP();
	    if (!is_true(cond)) /* jump if false */
		JUMP(READ_BYTES(bv, bc.numbytes_label));
	    else {
		bv += bc.numbytes_label;
	    }
	    free_var(cond);
	}
	break;
	
      case OP_JUMP:
	JUMP(READ_BYTES(bv, bc.numbytes_label));
	break;
	
      case OP_FOR_LIST:
	{
	    unsigned id = READ_BYTES(bv, bc.numbytes_var_name);
	    Label lab = READ_BYTES(bv, bc.numbytes_label);
	    Var count, list;
	    
	    count = TOP_RT_VALUE;  /* will be a number */
	    list = NEXT_TOP_RT_VALUE;  /* should be a list */
	    if (list.type != TYPE_LIST) {
		RAISE_ERROR(E_TYPE);
		free_var(POP());
		free_var(POP());
		JUMP(lab);
	    } else if (count.v.num > list.v.list[0].v.num /* size */) {
		free_var(POP());
		free_var(POP());
		JUMP(lab);
	    } else {
		free_var(RUN_ACTIV.rt_env[id]);
		RUN_ACTIV.rt_env[id] = var_ref(list.v.list[count.v.num]);
		count.v.num++; /* increment count */
		TOP_RT_VALUE = count;
	    }
	}
	break;
	
      case OP_FOR_RANGE:
	{
	    unsigned id = READ_BYTES(bv, bc.numbytes_var_name);
	    Label lab = READ_BYTES(bv, bc.numbytes_label);
	    Var from, to;
	    
	    to = TOP_RT_VALUE;
	    from = NEXT_TOP_RT_VALUE;
	    if ((to.type != TYPE_NUM && to.type != TYPE_OBJ)
		|| to.type != from.type) {
		RAISE_ERROR(E_TYPE);
		free_var(POP());
		free_var(POP());
		JUMP(lab);
	    } else if (to.type == TYPE_NUM
		       ? from.v.num > to.v.num
		       : from.v.obj > to.v.obj) {
		free_var(POP());
		free_var(POP());
		JUMP(lab);
	    } else {
		free_var(RUN_ACTIV.rt_env[id]);
		RUN_ACTIV.rt_env[id] = var_ref(from);
		if (to.type == TYPE_NUM)
		    from.v.num++;
		else
		    from.v.obj++;
		NEXT_TOP_RT_VALUE = from;
	    }
	}
	break;
	
      case OP_POP:
	free_var(POP());
	break;
	
      case OP_IMM:
	{
	    int slot = READ_BYTES(bv, bc.numbytes_literal); 
	    PUSH_DUP(RUN_ACTIV.prog->literals[slot]);
	}
	break;
	
      case OP_MAKE_EMPTY_LIST:
	{
	    Var list;

	    list = new_list(0);
	    PUSH(list);
	}
	break;
	
      case OP_LIST_ADD_TAIL:
	{
	    Var tail, list;

	    tail = POP();  /* whatever */
	    list = POP();  /* should be list */
	    if (list.type != TYPE_LIST) {
		free_var(list);
		free_var(tail);
		PUSH_ERROR(E_TYPE);
	    } else
		PUSH(listappend(list, tail));
	}
	break;
	
      case OP_LIST_APPEND:
	{
	    Var tail, list;
	    
	    tail = POP();  /* second, should be list */
	    list = POP();  /* first, should be list */
	    if (tail.type != TYPE_LIST  ||  list.type != TYPE_LIST) {
		free_var(tail);
		free_var(list);
		PUSH_ERROR(E_TYPE);
	    } else
		PUSH(listconcat(list, tail));
	}
	break;

      case OP_INDEXSET:
	{
	    Var value, index, list;

	    value = POP();  /* rhs value */
	    index = POP();  /* index, should be number */
	    list = POP();  /* lhs except last index, should be list or str */
	    /* whole thing should mean list[index] = value */
	    if ((list.type != TYPE_LIST && list.type != TYPE_STR)
		|| index.type != TYPE_NUM
		|| (list.type == TYPE_STR && value.type != TYPE_STR)) {
		free_var(value);
		free_var(index);
		free_var(list);
		PUSH_ERROR(E_TYPE);
	    } else if (index.v.num < 1
		       || (list.type == TYPE_LIST
			   && index.v.num > list.v.list[0].v.num /* size */)
		       || (list.type == TYPE_STR
			   && index.v.num > (int) strlen(list.v.str))) {
		free_var(value);
		free_var(index);
		free_var(list);
		PUSH_ERROR(E_RANGE);
	    } else if (list.type == TYPE_STR && strlen(value.v.str) != 1) {
		free_var(value);
		free_var(index);
		free_var(list);
		PUSH_ERROR(E_INVARG);
	    } else if (list.type == TYPE_LIST) {
		PUSH(listset(var_dup(list), value, index.v.num));
		free_var(list);
	    } else { /* TYPE_STR */
		char *tmp_str = str_dup(list.v.str);
		free_str(list.v.str);
		tmp_str[index.v.num - 1] = value.v.str[0];
		list.v.str = tmp_str;
		free_var(value);
		PUSH(list);
	    }
	}
	break;
	
      case OP_MAKE_SINGLETON_LIST:
	{
	    Var list;

	    list = new_list(1);
	    list.v.list[1] = POP();
	    PUSH(list);
	}
	break;
	
      case OP_CHECK_LIST_FOR_SPLICE:
	if (TOP_RT_VALUE.type != TYPE_LIST) {
	    free_var(POP());
	    PUSH_ERROR(E_TYPE);
	}
	/* no op if top-rt-stack is a list */
	break;
	
      case OP_PUT_TEMP:
	RUN_ACTIV.temp = var_ref(TOP_RT_VALUE);
	break;
	
      case OP_PUSH_TEMP:
	PUSH(RUN_ACTIV.temp);
	RUN_ACTIV.temp.type = TYPE_ANY;
	break;
	
      case OP_EQ:
      case OP_NE:
	{
	    Var rhs, lhs, ans;

	    rhs = POP();
	    lhs = POP();
	    ans.type = TYPE_NUM;
	    ans.v.num = (op == OP_EQ ? equality(rhs, lhs, 0)
			 : !equality(rhs, lhs, 0));
	    PUSH(ans);
	    free_var(rhs);
	    free_var(lhs);
	}
	break;
	
      case OP_GT:
      case OP_LT:
      case OP_GE:
      case OP_LE:
	{
	    Var rhs, lhs, ans;

	    rhs = POP();
	    lhs = POP();  /* should be same type as rhs and both non-list */
	    if (rhs.type != lhs.type || rhs.type == TYPE_LIST) {
		free_var(rhs);
		free_var(lhs);
		PUSH_ERROR(E_TYPE);
	    } else {
		int comparison;
		switch(rhs.type)  {
		  case TYPE_NUM:
		    comparison = compare_ints(lhs.v.num, rhs.v.num);
		    break;
		  case TYPE_OBJ:
		    comparison = compare_ints(lhs.v.obj, rhs.v.obj);
		    break;
		  case TYPE_ERR:
		    comparison = ((int) lhs.v.err) - ((int) rhs.v.err);
		    break;
		  case TYPE_STR:
		    comparison = mystrcasecmp(lhs.v.str, rhs.v.str);
		    break;
		  default:
		    errlogf("RUN: Impossible type in comparison: %d\n",
			   rhs.type);
		    comparison = 0;
		}	  
		ans.type = TYPE_NUM;
		switch (op) {
		  case OP_LT:
		    ans.v.num = (comparison < 0);
		    break;
		  case OP_LE:
		    ans.v.num = (comparison <= 0);
		    break;
		  case OP_GT:
		    ans.v.num = (comparison > 0);
		    break;
		  case OP_GE:
		    ans.v.num = (comparison >= 0);
		    break;
		  default:
		    errlogf("RUN: Imposible opcode in comparison: %d\n", op);
		    break;
		}
		PUSH(ans);
		free_var(rhs);
		free_var(lhs);
	    }
	}
	break;
	
      case OP_IN:
	{
	    Var lhs, rhs, ans;

	    rhs = POP(); /* should be list */
	    lhs = POP(); /* lhs, any type */
	    if (rhs.type != TYPE_LIST) {
		free_var(rhs);
		free_var(lhs);
		PUSH_ERROR(E_TYPE);
	    } else {
		ans.type = TYPE_NUM;
		ans.v.num = ismember(lhs, rhs, 0);
		PUSH(ans);
		free_var(rhs);
		free_var(lhs);
	    }
	}
	break;
	
      case OP_MULT:
      case OP_MINUS:
      case OP_DIV:
      case OP_MOD:
	{
	    Var lhs, rhs, ans;
	    
	    rhs = POP(); /* should be number */
	    lhs = POP(); /* should be number */
	    if (rhs.type != TYPE_NUM || lhs.type != TYPE_NUM) {
		free_var(rhs);
		free_var(lhs);
		PUSH_ERROR(E_TYPE);
	    } else if ((op == OP_DIV || op == OP_MOD) && rhs.v.num == 0) {
		free_var(rhs);
		free_var(lhs);
		PUSH_ERROR(E_DIV);
	    } else {
		ans.type = TYPE_NUM;
		switch (op) {
		  case OP_MULT: ans.v.num = lhs.v.num * rhs.v.num; break;
		  case OP_MINUS: ans.v.num = lhs.v.num - rhs.v.num; break;
		  case OP_DIV: ans.v.num = lhs.v.num / rhs.v.num; break;
		  case OP_MOD: ans.v.num = lhs.v.num % rhs.v.num; break;
		  default: errlogf("RUN: Impossible opcode in arith ops: %d\n",
				  op);
		    break;
		}
		PUSH(ans);
		free_var(rhs);
		free_var(lhs);
	    }
	}
	break;
	
      case OP_ADD:
	{
	    Var rhs, lhs, ans;
	    
	    rhs = POP(); /* should be number or string */
	    lhs = POP(); /* should be same type as rhs */
	    if (rhs.type != lhs.type ||
		(rhs.type != TYPE_NUM && rhs.type != TYPE_STR)) {
		free_var(rhs);
		free_var(lhs);
		PUSH_ERROR(E_TYPE);
	    } else {
		ans.type = rhs.type;
		if (rhs.type == TYPE_NUM)
		    ans.v.num = lhs.v.num + rhs.v.num;
		else {
		    char *str;
		    
		    str = mymalloc((strlen(rhs.v.str) +
				    strlen(lhs.v.str) + 1)
				   * sizeof(char), M_STRING);
		    sprintf(str, "%s%s", lhs.v.str, rhs.v.str);
		    ans.v.str = str;
		}
		PUSH(ans);
		free_var(rhs);
		free_var(lhs);
	    }
	}
	break;
	
      case OP_AND:
      case OP_OR:
	{
	    Var lhs;
	    unsigned lab = READ_BYTES(bv, bc.numbytes_label);
	    
	    lhs = TOP_RT_VALUE;
	    if (   (op == OP_AND && !is_true(lhs))
		|| (op == OP_OR  &&  is_true(lhs))) /* short-circuit */
		JUMP(lab);
	    else {
		free_var(POP());	
	    }
	}
	break;
	
      case OP_UNARY_MINUS:
      case OP_NOT:
	{
	    Var arg, ans;
	    
	    arg = POP(); /* should be number for OP_UNARY_MINUS,
			    any type for OP_NOT */
	    if (op == OP_UNARY_MINUS && arg.type != TYPE_NUM) {
		free_var(arg);
		PUSH_ERROR(E_TYPE);
	    } else {
		ans.type = TYPE_NUM;
		ans.v.num = (op == OP_NOT ? !is_true(arg) : -arg.v.num);
		PUSH(ans);
		free_var(arg);
	    }
	}
	break;

      case OP_REF:
	{
	    Var index, list;
	    
	    index = POP();  /* should be number */
	    list = POP();  /* should be list or string */
	    
	    if (index.type != TYPE_NUM ||
		(list.type != TYPE_LIST && list.type != TYPE_STR)) {
		free_var(index);
		free_var(list);
		PUSH_ERROR(E_TYPE);
	    }
	    else if (list.type == TYPE_LIST) {
		if (index.v.num <= 0 || index.v.num > list.v.list[0].v.num) {
		    free_var(index);
		    free_var(list);
		    PUSH_ERROR(E_RANGE);
		} else {
		    PUSH(var_ref(list.v.list[index.v.num]));
		    free_var(index);
		    free_var(list);
		}
	    } else { /* list.type == TYPE_STR */
		if (index.v.num <= 0
		    || index.v.num > (int) strlen(list.v.str)) {
		    free_var(index);
		    free_var(list);
		    PUSH_ERROR(E_RANGE);
		} else {
		    PUSH(strget(list, index));
		    free_var(index);
		    free_var(list);
		}
	    }
	}
	break;

      case OP_PUSH_REF:
	{
	    Var index, list;
	    
	    index = TOP_RT_VALUE;
	    list = NEXT_TOP_RT_VALUE;

	    if (index.type != TYPE_NUM || list.type != TYPE_LIST) {
		PUSH_ERROR(E_TYPE);
	    } else if (index.v.num <= 0 ||
		       index.v.num > list.v.list[0].v.num) {
		PUSH_ERROR(E_RANGE);
	    } else 
		PUSH(var_ref(list.v.list[index.v.num]));
	}
	break;

      case OP_RANGE_REF:
	{
	    Var base, from, to;
	    
	    to = POP(); /* should be number */
	    from = POP(); /* should be number */
	    base = POP(); /* should be list or string */
	    
	    if ((base.type != TYPE_LIST && base.type != TYPE_STR) 
		|| to.type != TYPE_NUM || from.type != TYPE_NUM) {
		free_var(to);
		free_var(from);
		PUSH_ERROR(E_TYPE);
	    } else {
		int len = (base.type == TYPE_STR ? strlen(base.v.str) 
			   : base.v.list[0].v.num);
		if (from.v.num <= to.v.num
		    && (from.v.num <= 0 || from.v.num > len
			|| to.v.num <= 0 || to.v.num > len)) {
		    free_var(to);
		    free_var(from);
		    free_var(base);
		    PUSH_ERROR(E_RANGE);
		} else {
		    PUSH((base.type == TYPE_STR 
			  ? substr(base, from.v.num, to.v.num) 
			  : sublist(base, from.v.num, to.v.num)));
		    /* base freed by substr/sublist */
		    free_var(from);
		    free_var(to);
		}
	    }
	}
	break;
	
      case OP_G_PUT:
	{
	    unsigned id = READ_BYTES(bv, bc.numbytes_var_name);
	    free_var(RUN_ACTIV.rt_env[id]);
	    RUN_ACTIV.rt_env[id] = var_ref(TOP_RT_VALUE);
	}
	break;
	
      case OP_G_PUSH:
	{
	    Var value;
	    
	    value = RUN_ACTIV.rt_env[READ_BYTES(bv, bc.numbytes_var_name)];
	    if (value.type == TYPE_ANY)
		PUSH_ERROR(E_VARNF)
	    else
		PUSH_DUP(value);
	}
	break;
	
      case OP_GET_PROP:
	{
	    Var propname, obj, prop;
	    
	    propname = POP(); /* should be string */
	    obj = POP(); /* should be objid */
	    if (propname.type != TYPE_STR || obj.type != TYPE_OBJ) {
		free_var(propname);
		free_var(obj);
		PUSH_ERROR(E_TYPE);
	    } else if (!valid(obj.v.obj)) {
		free_var(propname);
		free_var(obj);
		PUSH_ERROR(E_INVIND);
	    } else {
		enum error err = get_prop_val(obj.v.obj, propname.v.str, &prop,
					  RUN_ACTIV.progr, 0);
		free_var(propname);
		free_var(obj);
		if (err == E_NONE)
		    PUSH(prop);
		else
		    PUSH_ERROR(err);
	    }
	}
	break;
	
      case OP_PUSH_GET_PROP:
	{
	    Var propname, obj, prop;
	    
	    propname = TOP_RT_VALUE;
	    obj = NEXT_TOP_RT_VALUE;
	    if (propname.type != TYPE_STR || obj.type != TYPE_OBJ) {
		PUSH_ERROR(E_TYPE);
	    } else if (!valid(obj.v.obj)) {
		PUSH_ERROR(E_INVIND);
	    } else {
		enum error err = get_prop_val(obj.v.obj, propname.v.str, &prop,
					  RUN_ACTIV.progr, 0);
		if (err == E_NONE)
		    PUSH(prop);
		else
		    PUSH_ERROR(err);
	    }
	}
	break;
	
      case OP_PUT_PROP:
	{
	    Var obj, propname, rhs;
	    
	    rhs = POP();      /* any type */
	    propname = POP(); /* should be string */
	    obj = POP();      /* should be objid */
	    if (obj.type != TYPE_OBJ || propname.type != TYPE_STR) {
		free_var(rhs);
		free_var(propname);
		free_var(obj);
		PUSH_ERROR(E_TYPE);
	    } else if (!valid(obj.v.obj)) {
		free_var(rhs);
		free_var(propname);
		free_var(obj);
		PUSH_ERROR(E_INVIND);
	    } else {
		enum error err = set_prop_val(obj.v.obj, propname.v.str, rhs,
					  RUN_ACTIV.progr); /* rhs !consumed */
		free_var(propname);
		free_var(obj);
		if (err == E_NONE)
		    PUSH(rhs);
		else 
		    PUSH_ERROR(err);
	    }
	}
	break;
	
      case OP_FORK:
      case OP_FORK_WITH_ID:
	{
	    Var time;
	    unsigned id = 0, f_index;

	    time = POP();
	    f_index = READ_BYTES(bv, bc.numbytes_fork);
	    if (op == OP_FORK_WITH_ID)
		id = READ_BYTES(bv, bc.numbytes_var_name);
	    if (time.type != TYPE_NUM) {
		free_var(time);
		RAISE_ERROR(E_TYPE);
	    } else if (time.v.num < 0) {
		free_var(time);
		RAISE_ERROR(E_INVARG);
	    } else {
		Var    *copied_rt_env = copy_rt_env(RUN_ACTIV.rt_env,
					  RUN_ACTIV.prog->num_var_names);
		Var	task_id;

		task_id = enqueue_forked_task(RUN_ACTIV.prog, RUN_ACTIV,
					      copied_rt_env, f_index,
					      time.v.num);
		if (task_id.type == TYPE_ERR) {
		    free_rt_env(copied_rt_env, RUN_ACTIV.prog->num_var_names);
		    RAISE_ERROR(task_id.v.err);
		} else {
		    add_ref_to_program(RUN_ACTIV.prog);
		    if (op == OP_FORK_WITH_ID) {
			free_var(RUN_ACTIV.rt_env[id]);
			RUN_ACTIV.rt_env[id] = task_id;
			free_var(copied_rt_env[id]);
			copied_rt_env[id] = task_id;
		    }
		}
	    }
	}
	break;
	
      case OP_CALL_VERB: 
	{
	    enum error err;
	    Var args, verb, obj;
	    
	    args = POP(); /* args, should be list */
	    verb = POP(); /* verbname, should be string */
	    obj = POP(); /* objid, should be obj */

	    if (args.type != TYPE_LIST || verb.type != TYPE_STR
		|| obj.type != TYPE_OBJ)
		err = E_TYPE;
	    else if (!valid(obj.v.obj))
		err = E_INVIND;
	    else {
		STORE_STATE_VARIABLES();
		err = call_verb(obj.v.obj, verb.v.str, args, 0);
		/* if there is no error, RUN_ACTIV is now the CALLEE's.
		   args will be consumed in the new rt_env */
		/* if there is an error, then RUN_ACTIV is unchanged, and
		   args is not consumed in this case */
		LOAD_STATE_VARIABLES();
	    }
	    free_var(obj);
	    free_var(verb);

	    if (err != E_NONE) { /* there is an error, RUN_ACTIV unchanged, 
				    args must be freed */
		free_var(args);
		PUSH_ERROR(err);
	    }
	}
	break; 
	
      case OP_RETURN:
      case OP_RETURN0:
      case OP_DONE:
	{
	    void *bi_func_data = 0;
	    int func_pc;
	    Var ret_val;
	  
	    if (op == OP_RETURN) {
		ret_val = POP();
	    } else {
		ret_val.type = TYPE_NUM;
		ret_val.v.num = 0;
	    }
	    func_pc = RUN_ACTIV.bi_func_pc;
	    if (func_pc) {
		func_id = RUN_ACTIV.bi_func_id;
		bi_func_data = RUN_ACTIV.bi_func_data;
	    }
	  
	    STORE_STATE_VARIABLES();
	    free_activation(RUN_ACTIV, 0); /* 0 == dont free bi_func_data */
	  
	    if (top_activ_stack == 0) { /* root activation finished */
		if (aborting_task)
		    return OUTCOME_ABORTED;
		else {
		    if (result)
			*result = ret_val;
		    else
			free_var(ret_val);
		    return OUTCOME_DONE;
		}
	    }
	  
	    top_activ_stack--;
	    LOAD_STATE_VARIABLES();
	  
	    if (func_pc == 0)
		PUSH(ret_val);
	    else {
		STORE_STATE_VARIABLES();
		p = call_bi_func(func_id, ret_val, func_pc, RUN_ACTIV.progr, 
				 bi_func_data);
		LOAD_STATE_VARIABLES();
		goto handle_bi_outcome;
	    }
	}
	break;
	
      case OP_BI_FUNC_CALL:
	{
	    Var args;
	    
	    func_id = READ_BYTES(bv, 1); /* 1 == numbytes of func_id */
	    args = POP();	/* should be list */
	    if (args.type != TYPE_LIST) {
		free_var(args);
		PUSH_ERROR(E_TYPE);
	    } else {
		STORE_STATE_VARIABLES();
		p = call_bi_func(func_id, args, 1, RUN_ACTIV.progr, 0); 
		LOAD_STATE_VARIABLES();

	      handle_bi_outcome:
		switch (p.kind) {
		  case BI_RETURN:
		    PUSH(p.u.ret);
		    break;
		  case BI_RAISE:
		    if (!aborting_task)
			PUSH_ERROR(p.u.err);
		    break;
		  case BI_CALL:
		    /* another activ has been pushed onto activ_stack */
		    RUN_ACTIV.bi_func_id = func_id;
		    RUN_ACTIV.bi_func_data = p.u.call.data;
		    RUN_ACTIV.bi_func_pc = p.u.call.pc;
		    break;
		  case BI_SUSPEND:
		    {
			vm 		the_vm = new_vm(top_activ_stack + 1);
			int		i;
			enum error 	e;

			the_vm->max_stack_size = max_stack_size;
			the_vm->top_activ_stack = top_activ_stack;
			the_vm->root_activ_vector = root_activ_vector;
			the_vm->func_id = 0; /* shouldn't need func_id; */
			for (i = 0; i <= top_activ_stack; i++)
			    the_vm->activ_stack[i] = activ_stack[i];

			e = (*p.u.susp.proc)(the_vm, p.u.susp.data);
			if (e == E_NONE)
			    return OUTCOME_BLOCKED;
			else {
			    free_vm(the_vm, 0);
			    PUSH_ERROR(e);
			}
		    }
		}
	    }
        }
	break;

      default:
	if (IS_PUSH_n(op)) {
	    Var value;
	    value = RUN_ACTIV.rt_env[PUSH_n_INDEX(op)];
	    if (value.type == TYPE_ANY) {
		free_var(value);
		PUSH_ERROR(E_VARNF);
	    } else
		PUSH_DUP(value);
	} else if (IS_PUT_n(op)) {
	    free_var(RUN_ACTIV.rt_env[PUT_n_INDEX(op)]);
	    RUN_ACTIV.rt_env[PUT_n_INDEX(op)] = var_ref(TOP_RT_VALUE);
	} else if (IS_OPTIM_NUM_OPCODE(op)) {
	    Var value;
	    value.type = TYPE_NUM;
	    value.v.num = OPCODE_TO_OPTIM_NUM(op);
	    PUSH(value);
	}
	break;
    }
  }
}


/**** manipulating data of task ****/

static int	timeouts_enabled = 1;	/* set to 0 in debugger to disable
					   timeouts */

static void
task_timeout(Timer_ID id, Timer_Data data)
{
    task_timed_out = timeouts_enabled;
}

static Timer_ID
setup_task_execution_limits(int seconds, int ticks)
{
    task_alarm_id = set_virtual_timer(seconds < 1 ? 1 : seconds,
				      task_timeout, 0);
    task_timed_out = aborting_task = 0;
    ticks_remaining = (ticks < 100 ? 100 : ticks);
    return task_alarm_id;
}
  

void
set_task_kind(task_kind kind)
{
    current_task_kind = kind;
}

static void
check_activ_stack_size(unsigned max)
{
    if (max_stack_size != max) {
	if (activ_stack)
	    myfree(activ_stack, M_VM);

	activ_stack = mymalloc(sizeof(activation) * max, M_VM);
	max_stack_size = max;
    }
}

unsigned
current_max_stack_size(void)
{
    int	max = server_int_option("max_stack_depth", DEFAULT_MAX_STACK_DEPTH);

    if (max < DEFAULT_MAX_STACK_DEPTH)
	max = DEFAULT_MAX_STACK_DEPTH;

    return max;
}

/**** There are two methods of starting a new task:
   (1) Create a new one
   (2) Resume an old one  */

/* procedure to create a new task */

static enum outcome
do_task(Program *prog, unsigned which_vector, Var *result)
{ 	/* which vector determines the vector for the root_activ.
	   a forked task can also have which_vector == MAIN_VECTOR.
	   this happens iff it is recovered from a read from disk,
	   because in that case the forked statement is parsed as 
	   the main vector */
    enum outcome	ret;
    int			forked = (current_task_kind == TASK_FORKED);

    setup_task_execution_limits(forked ? server_int_option("bg_seconds",
							   DEFAULT_BG_SECONDS)
				       : server_int_option("fg_seconds",
							   DEFAULT_FG_SECONDS),
				forked ? server_int_option("bg_ticks",
							   DEFAULT_BG_TICKS)
				       : server_int_option("fg_ticks",
							   DEFAULT_FG_TICKS));

    RUN_ACTIV.prog = prog;
    add_ref_to_program(prog);

    root_activ_vector = which_vector; /* main or which of the forked */
    RUN_ACTIV.top_rt_stack = RUN_ACTIV.base_rt_stack
	= mymalloc ((which_vector == MAIN_VECTOR
		     ? prog->main_vector.max_stack
		     : prog->fork_vectors[which_vector].max_stack) *
		    sizeof(Var), M_RT_STACK);

    RUN_ACTIV.pc = 0;
    RUN_ACTIV.error_pc = 0;
    RUN_ACTIV.bi_func_pc = 0;
    RUN_ACTIV.temp.type = TYPE_ANY;

    /* actual running */
    interpreter_is_running = 1;
    ret = run(E_NONE, result);
    interpreter_is_running = 0;
    task_timed_out = 0;

    (void) cancel_timer(task_alarm_id);

    return ret;
}

/* procedure to resume an old task */

enum outcome
resume_from_previous_vm(vm the_vm, Var v, Var *result)
{
    enum outcome	ret;
    int			i;

    setup_task_execution_limits(server_int_option("bg_seconds",
						  DEFAULT_BG_SECONDS),
				server_int_option("bg_ticks",
						  DEFAULT_BG_TICKS));
  
    check_activ_stack_size(the_vm->max_stack_size);
    top_activ_stack = the_vm->top_activ_stack;
    root_activ_vector = the_vm->root_activ_vector;
    for (i = 0; i <= top_activ_stack; i++)
	activ_stack[i] = the_vm->activ_stack[i];

    free_vm(the_vm, 0);

    interpreter_is_running = 1;
    if (v.type == TYPE_ERR)
	ret = run(v.v.err, result);
    else {
	/* PUSH(v) */
	*((RUN_ACTIV.top_rt_stack)++) = var_ref(v);

	ret = run(E_NONE, result);
    }
    interpreter_is_running = 0;
    task_timed_out = 0;

    (void) cancel_timer(task_alarm_id);
    return ret;
}


/*** external functions ***/



enum outcome
do_server_task(Objid this, const char *verb, Var args, Objid vloc,
	       const char *verbname, Program *prog, Objid progr, int debug,
	       Objid player, const char *argstr, Var *result)
/* suspending should be dis-abled in do_server_task */
{
    Var *env;
    
    current_task_kind = TASK_INPUT;
    check_activ_stack_size(current_max_stack_size());
    top_activ_stack = 0;

    RUN_ACTIV.rt_env = env = new_rt_env(prog->num_var_names);
    RUN_ACTIV.this = this;
    RUN_ACTIV.player = player;
    RUN_ACTIV.progr = progr;
    RUN_ACTIV.vloc = vloc;
    RUN_ACTIV.verb = str_dup(verb);
    RUN_ACTIV.verbname = str_ref(verbname);
    RUN_ACTIV.debug = debug;
    fill_in_rt_consts(env);
    set_rt_env_obj(env, SLOT_PLAYER, player);
    set_rt_env_obj(env, SLOT_CALLER, -1);
    set_rt_env_obj(env, SLOT_THIS, this);
    set_rt_env_obj(env, SLOT_DOBJ, NOTHING);
    set_rt_env_obj(env, SLOT_IOBJ, NOTHING);
    set_rt_env_str(env, SLOT_DOBJSTR, str_dup(""));
    set_rt_env_str(env, SLOT_IOBJSTR, str_dup(""));
    set_rt_env_str(env, SLOT_ARGSTR, str_dup(argstr));
    set_rt_env_str(env, SLOT_PREPSTR, str_dup(""));
    set_rt_env_str(env, SLOT_VERB, str_ref(RUN_ACTIV.verb));
    set_rt_env_var(env, SLOT_ARGS, args);

    return do_task(prog, MAIN_VECTOR, result);
}

enum outcome
do_input_task(Program *prog, Parse_Info *pi, Var *result)
{
    Var *env;

    current_task_kind = TASK_INPUT;
    check_activ_stack_size(current_max_stack_size());
    top_activ_stack = 0;

    RUN_ACTIV.rt_env = env = new_rt_env(prog->num_var_names);
    RUN_ACTIV.this = pi->this;
    RUN_ACTIV.player = pi->player;
    RUN_ACTIV.progr = pi->progr;
    RUN_ACTIV.vloc = pi->vloc;
    RUN_ACTIV.verb = str_ref(pi->verb);
    RUN_ACTIV.verbname = str_ref(pi->verbname);
    RUN_ACTIV.debug = pi->debug;
    fill_in_rt_consts(env);
    set_rt_env_obj(env, SLOT_PLAYER, pi->player);
    set_rt_env_obj(env, SLOT_CALLER, pi->caller);
    set_rt_env_obj(env, SLOT_THIS, pi->this);
    set_rt_env_obj(env, SLOT_DOBJ, pi->dobj);
    set_rt_env_obj(env, SLOT_IOBJ, pi->iobj);
    set_rt_env_str(env, SLOT_DOBJSTR, str_ref(pi->dobjstr));
    set_rt_env_str(env, SLOT_IOBJSTR, str_ref(pi->iobjstr));
    set_rt_env_str(env, SLOT_ARGSTR, str_ref(pi->argstr));
    set_rt_env_str(env, SLOT_PREPSTR, str_ref(pi->prepstr));
    set_rt_env_str(env, SLOT_VERB, str_ref(pi->verb));
    set_rt_env_var(env, SLOT_ARGS, var_ref(pi->args));

    return do_task(prog, MAIN_VECTOR, result);
}

enum outcome
do_forked_task(Program *prog, Var *rt_env, activation a, unsigned f_id,
	       Var *result)
{    
    current_task_kind = TASK_FORKED;	
    check_activ_stack_size(current_max_stack_size());
    top_activ_stack = 0;

    RUN_ACTIV = a;
    RUN_ACTIV.rt_env = rt_env;
    
    return do_task(prog, f_id, result);
}

/* this is called from bf_eval to set up stack for an eval call */

int
setup_activ_for_eval(Program *prog)
{
    Var *env;
    if (!push_activation())
	return 0;
    
    RUN_ACTIV.prog = prog;
    
    RUN_ACTIV.rt_env = env = new_rt_env(prog->num_var_names);
    fill_in_rt_consts(env);
    set_rt_env_obj(env, SLOT_PLAYER, CALLER_ACTIV.player);
    set_rt_env_obj(env, SLOT_CALLER, CALLER_ACTIV.this);
    set_rt_env_obj(env, SLOT_THIS, NOTHING);
    set_rt_env_obj(env, SLOT_DOBJ, NOTHING);
    set_rt_env_obj(env, SLOT_IOBJ, NOTHING);
    set_rt_env_str(env, SLOT_DOBJSTR, str_dup(""));
    set_rt_env_str(env, SLOT_IOBJSTR, str_dup(""));
    set_rt_env_str(env, SLOT_ARGSTR, str_dup(""));
    set_rt_env_str(env, SLOT_PREPSTR, str_dup(""));
    set_rt_env_str(env, SLOT_VERB, str_dup(""));
    set_rt_env_var(env, SLOT_ARGS, new_list(0));
    
    RUN_ACTIV.this = NOTHING;
    RUN_ACTIV.player = CALLER_ACTIV.player;
    RUN_ACTIV.progr = CALLER_ACTIV.progr;
    RUN_ACTIV.vloc = NOTHING;
    RUN_ACTIV.verb = str_dup("");
    RUN_ACTIV.verbname = str_dup("Input to EVAL");
    RUN_ACTIV.debug = 1;
    RUN_ACTIV.top_rt_stack = RUN_ACTIV.base_rt_stack 
	= mymalloc (RUN_ACTIV.prog->main_vector.max_stack * sizeof(Var), 
		    M_RT_STACK);
    RUN_ACTIV.pc = 0;
    RUN_ACTIV.error_pc = 0;
    RUN_ACTIV.temp.type = TYPE_ANY;
    
    return 1;
}

void
abort_running_task(void)
{
    aborting_task = 1;
}

/**** built in functions ****/

static package
bf_suspend(Var arglist, Byte next, void *vdata, Objid progr)
{
    static int	seconds;

    seconds = arglist.v.list[1].v.num;
    free_var(arglist);

    if (seconds < 0) 
	return make_error_pack(E_INVARG);
    else
	return make_suspend_pack(enqueue_suspended_task, &seconds);
}

static package
bf_read(Var arglist, Byte next, void *vdata, Objid progr)
{
    int			argc = arglist.v.list[0].v.num;
    static Objid	connection;
    int			non_blocking = (argc >= 2
					&& is_true(arglist.v.list[2]));

    if (argc >= 1)
	connection = arglist.v.list[1].v.obj;
    else
	connection = activ_stack[0].player;
    free_var(arglist);

    /* Permissions checking */
    if (argc >= 1) {
	if (!is_wizard(progr)
	    &&  (!valid(connection)  ||  progr != obj_owner(connection)))
	    return make_error_pack(E_PERM);
    } else {
	if (!is_wizard(progr)
	    ||  last_input_task_id(connection) != current_task_id)
	    return make_error_pack(E_PERM);
    }

    if (non_blocking) {
	Var		r;

	r = read_input_now(connection);
	if (r.type == TYPE_ERR)
	    return make_error_pack(r.v.err);
	else
	    return make_var_pack(r);
    }

    return make_suspend_pack(make_reading_task, &connection);
}

static package
bf_seconds_left(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    r.type = TYPE_NUM;
    r.v.num = timer_wakeup_interval(task_alarm_id);
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_ticks_left(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var r;
    r.type = TYPE_NUM;
    r.v.num = ticks_remaining;
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_pass(Var arglist, Byte next, void *vdata, Objid progr)
{
    if (next == 1) {
	enum error e = call_verb(RUN_ACTIV.this, RUN_ACTIV.verb, arglist, 1);
	
	if (e == E_NONE) 
	    return make_call_pack(2, 0);
	
	free_var(arglist);
	return make_error_pack(e);
    } else
	return make_var_pack(arglist);
}

static package
bf_set_task_perms(Var arglist, Byte next, void *vdata, Objid progr) 
{ /* (player) */
    /* warning!!  modifies top activation */
    Objid oid = arglist.v.list[1].v.obj;
    
    free_var(arglist);
    
    if (progr != oid && !is_wizard(progr)) 
	return make_error_pack(E_PERM);
    
    RUN_ACTIV.progr = oid;
    return no_var_pack();
}

static package
bf_caller_perms(Var arglist, Byte next, void *vdata, Objid progr) 
{ /* () */
    Var r;
    r.type = TYPE_OBJ;
    if (top_activ_stack == 0)
	r.v.obj = NOTHING;
    else
	r.v.obj = activ_stack[top_activ_stack - 1].progr;
    free_var(arglist);
    return make_var_pack(r);
}

static package
bf_callers(Var arglist, Byte next, void *vdata, Objid progr)
{
    Var 	r;
    int		i;
    
    r = new_list(top_activ_stack);
    for (i = top_activ_stack - 1; i >= 0; i--) {
	Var	v;
	
	v = r.v.list[top_activ_stack - i] = new_list(5);
	v.v.list[1].type = TYPE_OBJ;
	v.v.list[1].v.obj = activ_stack[i].this;
	v.v.list[2].type = TYPE_STR;
	v.v.list[2].v.str = str_ref(activ_stack[i].verb);
	v.v.list[3].type = TYPE_OBJ;
	v.v.list[3].v.obj = activ_stack[i].progr;
	v.v.list[4].type = TYPE_OBJ;
	v.v.list[4].v.obj = activ_stack[i].vloc;
	v.v.list[5].type = TYPE_OBJ;
	v.v.list[5].v.obj = activ_stack[i].player;
    }
    
    free_var(arglist);
    return make_var_pack(r);
}

void 
register_execute(void) /* declared in bf_register.h */
{
    register_function("suspend", 1, 1, bf_suspend, TYPE_NUM);
    register_function("read", 0, 2, bf_read, TYPE_OBJ, TYPE_ANY);
    
    (void) register_function("seconds_left", 0, 0, bf_seconds_left);
    (void) register_function("ticks_left", 0, 0, bf_ticks_left);
    (void) register_function("pass", 0, -1, bf_pass);
    (void) register_function("set_task_perms", 1, 1, bf_set_task_perms,
			     TYPE_OBJ);
    (void) register_function("caller_perms", 0, 0, bf_caller_perms);
    (void) register_function("callers", 0, 0, bf_callers);
}

/**** storing to/loading from database ****/

void
write_activ(activation a)
{
    register Var *rt_stack_runner = a.base_rt_stack;

    write_program(a.prog);
    write_rt_env(a.prog->var_names, a.rt_env, a.prog->num_var_names);
    
    db_printf("%d rt_stack slots in use\n", a.top_rt_stack - a.base_rt_stack);

    for (; rt_stack_runner != a.top_rt_stack; rt_stack_runner++) 
	write_var(*rt_stack_runner);

    write_activ_as_pi(a);
    write_var(a.temp);
    
    db_printf("%u %u %u\n", a.pc, a.bi_func_pc, a.error_pc);
    if (a.bi_func_pc != 0) {
	write_string(name_func_by_num(a.bi_func_id));
	write_bi_func_data(a.bi_func_data, a.bi_func_id);
    }
}

static int
check_pc_validity(Program *prog, int which_vector, unsigned pc)
{
    Bytecodes  *bc = (which_vector == -1
		      ? &prog->main_vector
		      : &prog->fork_vectors[which_vector]);

    /* Current insn must be call to verb or built-in function like eval(),
     * move(), pass(), or suspend().
     */
    return (pc < bc->size
	    && (bc->vector[pc - 1] == OP_CALL_VERB
		|| bc->vector[pc - 2] == OP_BI_FUNC_CALL));
}

int
read_activ(FILE *f, activation *a, int which_vector)
{
    Var        *old_rt_env;
    const char **old_names;
    int 	old_size, stack_in_use;
    unsigned	i;
    activation  sub_a;
    const char *func_name;
    int		max_stack;
    char	c;

    if (read_program(f, &(a->prog)) < 0) {
	errlogf("READ_ACTIV: Bad program\n");
	return -1;
    }

    if (read_rt_env(f, &old_names, &old_rt_env, &old_size) < 0) {
	errlogf("READ_ACTIV: Bad env\n");
	return -1;
    }

    a->rt_env = reorder_rt_env(old_rt_env, old_names, old_size, a->prog);

    max_stack = (which_vector == MAIN_VECTOR
		 ? a->prog->main_vector.max_stack
		 : a->prog->fork_vectors[which_vector].max_stack);
    a->base_rt_stack = mymalloc (max_stack * sizeof(Var), M_RT_STACK);

    if (fscanf(f, "%d rt_stack slots in use\n", &stack_in_use) != 1) {
	errlogf("READ_ACTIV: Bad stack_in_use number\n");
	return -1;
    }

    a->top_rt_stack = a->base_rt_stack;
    for (i = 0; i < stack_in_use; i++) 
	*(a->top_rt_stack++) = read_var(f);
    
    if (read_activ_as_pi(f, &sub_a) < 0) {
	errlogf("READ_ACTIV: Bad activ. stack_in_use = %d\n", stack_in_use);
	return -1;
    }
    a->this = sub_a.this;
    a->player = sub_a.player;
    a->progr = sub_a.progr;
    a->vloc = sub_a.vloc;
    a->debug = sub_a.debug;
    a->verb = sub_a.verb;
    a->verbname = sub_a.verbname;

    a->temp = read_var(f);

    if (fscanf(f, "%u %u", &(a->pc), &i) != 2) {
	errlogf("READ_ACTIV: bad pc, next. stack_in_use = %d\n", stack_in_use);
	return -1;
    }
    
    if (fscanf(f, "%c", &c) != 1) {
	errlogf("READ_ACTIV_ERROR_PC: no next character.\n");
	return -1;
    }
    if (c == '\n')
	a->error_pc = a->pc;
    else
	if (fscanf(f, "%u\n", &(a->error_pc)) != 1) {
	    errlogf("READ_ACTIV_ERROR_PC: no error pc.\n");
	    return -1;
	}

    if (!check_pc_validity(a->prog, which_vector, a->pc)) {
	errlogf("READ_ACTIV: Bad PC for suspended task.\n");
	return -1;
    }
    
    a->bi_func_pc = i;  /* should have written (Byte) bi_func_pc as a %c...?
			   in any case read it as %u into &(a->bi_func_pc)
			   does not work -- hence the hack with &i
			   above... sigh! */

    if (a->bi_func_pc != 0) {
	func_name = read_string(f);
	if ((a->bi_func_id = number_func_by_name(func_name)) == FUNC_NOT_FOUND
	    || read_bi_func_data(f, a->bi_func_id, &(a->bi_func_data)) < 0) {
	    errlogf("READ_ACTIV: Bad function. stack_in_use %d, pc %d, next %d\n",
		   stack_in_use, a->pc, a->bi_func_pc);
	    return -1;
	}
    }
    return 0;
}

char rcsid_execute[] = "$Id: execute.c,v 1.26 1992/10/23 23:03:47 pavel Exp $";

/* $Log: execute.c,v $
 * Revision 1.26  1992/10/23  23:03:47  pavel
 * Added copyright notice.
 *
 * Revision 1.25  1992/10/23  19:25:00  pavel
 * Eliminated all uses of the useless macro NULL..
 *
 * Revision 1.24  1992/10/21  03:02:35  pavel
 * Converted to use new automatic configuration system.
 *
 * Revision 1.23  1992/10/17  20:28:48  pavel
 * Global rename of strdup->str_dup, strref->str_ref, vardup->var_dup, and
 * varref->var_ref.
 * Added some (int) casts to placate over-protective compilers.
 * Changed return-type of read_activ() from char to int, for systems that use
 * unsigned chars.
 *
 * Revision 1.22  1992/09/25  21:11:36  pjames
 * Made use of the error_pc field in an activation to report correct
 * error tracebacks from verbs called by builtin verbs.
 *
 * Revision 1.21  1992/09/24  16:43:10  pavel
 * Exported `task_timed_out' for use by long-running built-in functions that
 * should abort if the timer goes off.
 *
 * Revision 1.20  1992/09/14  18:40:02  pjames
 * Updated #includes.  Moved rcsid to bottom.
 *
 * Revision 1.19  1992/09/14  17:45:17  pjames
 * Moved db_modification code to db modules.
 *
 * Revision 1.18  1992/09/08  22:07:06  pjames
 * Changed `register_bf_execute()' to `register_execute()'.
 * Changed range checking for subrange assigments.
 *
 * Revision 1.17  1992/09/04  22:42:05  pavel
 * Fixed some picky ANSI C problems with (const char *)'s.
 *
 * Revision 1.16  1992/09/03  16:29:44  pjames
 * Minor change.
 *
 * Revision 1.15  1992/09/02  18:41:17  pavel
 * Added mechanism for resuming tasks with errors.
 *
 * Revision 1.14  1992/09/01  06:24:01  pavel
 * Fixed minor storage leak just introduced.
 *
 * Revision 1.13  1992/09/01  05:30:57  pavel
 * Added a call to vardup() for first argument to listset().
 *
 * Revision 1.12  1992/08/31  22:29:01  pjames
 * Changed some `char *'s to `const char *' and fixed code accordingly.
 *
 * Revision 1.11  1992/08/28  23:23:03  pjames
 * Added support for EXTENDED instructions.
 * Added OP_RANGESET interpreting.
 * Changed OP_LISTSET to OP_INDEXSET, and added code for indexed string
 * assigments.
 *
 * Revision 1.10  1992/08/28  16:30:45  pjames
 * Changed myfree(*, M_STRING) to free_str(*).
 * Changed vardup to varref.
 * Changed some strref's to strdup.
 *
 * Revision 1.9  1992/08/12  02:01:46  pjames
 * Gave a variable an initial value to get rid on a bogus compiler
 * warning.
 *
 * Revision 1.8  1992/08/12  01:48:03  pjames
 * Fixed string freeing/copying to/from activation/forked_tasks.
 *
 * Revision 1.7  1992/08/11  17:30:44  pjames
 * Removed last read/write of Parse_Info, fixed a bug in a macro
 * (commented back in a line), and fixed call_verb (to not run the
 * program if it doesn't exist).
 *
 * Revision 1.6  1992/08/10  17:47:00  pjames
 * Moved several functions to eval_env.c and eval_vm.c.  Moved built in
 * functions which modified vm to this module.  Move property functions
 * to property.c.  Updated #includes.  Modified run() to be cleaner.
 * Used new built in function registration methods.  Removed uses of
 * Parse_Info field in activations.
 *
 * Revision 1.5  1992/08/01  01:14:52  pavel
 * Minor fixes...
 *
 * Revision 1.4  1992/07/30  21:18:48  pjames
 * Updated print_error_backtrace to print line numbers for each frame on
 * the stack.
 *
 * Revision 1.3  1992/07/27  18:14:17  pjames
 * Changed name of ct_env to var_names, const_env to literals, and
 * f_vectors to fork_vectors.
 *
 * Revision 1.2  1992/07/21  00:02:07  pavel
 * Added rcsid_<filename-root> declaration to hold the RCS ident. string.
 *
 * Revision 1.1  1992/07/20  23:23:12  pavel
 * Initial RCS-controlled version.
 */
